@@ -45,6 +45,14 @@
 #include <time.h>
 
 #include <linux/ioctl.h>
+
+#if HAVE_NETLINK
+#include "nbd-netlink.h"
+#include <netlink/netlink.h>
+#include <netlink/genl/genl.h>
+#include <netlink/genl/ctrl.h>
+#endif
+
 #define MY_NAME "nbd_client"
 #include "cliserv.h"
 
@@ -58,6 +66,123 @@
 
 #define NBDC_DO_LIST 1
 
+#if HAVE_NETLINK
+static int callback(struct nl_msg *msg, void *arg) {
+	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
+	struct nlattr *msg_attr[NBD_ATTR_MAX + 1];
+	int ret;
+	uint32_t index;
+
+	ret = nla_parse(msg_attr, NBD_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+			genlmsg_attrlen(gnlh, 0), NULL);
+	if (ret)
+		err("Invalid response from the kernel\n");
+	if (!msg_attr[NBD_ATTR_INDEX])
+		err("Did not receive index from the kernel\n");
+	index = nla_get_u32(msg_attr[NBD_ATTR_INDEX]);
+	printf("Connected /dev/nbd%d\n", (int)index);
+	return NL_OK;
+}
+
+static struct nl_sock *get_nbd_socket(int *driver_id) {
+	struct nl_sock *socket;
+
+	socket = nl_socket_alloc();
+	if (!socket)
+		err("Couldn't allocate netlink socket\n");
+
+	if (genl_connect(socket))
+		err("Couldn't connect to the generic netlink socket\n");
+	*driver_id = genl_ctrl_resolve(socket, "nbd");
+	if (*driver_id < 0)
+		err("Couldn't resolve the nbd netlink family, make sure the nbd module is loaded and your nbd driver supports the netlink interface.\n");
+	return socket;
+}
+
+static void netlink_configure(int index, int *sockfds, int num_connects,
+			      u64 size64, int blocksize, uint16_t flags,
+			      int timeout) {
+	struct nl_sock *socket;
+	struct nlattr *sock_attr;
+	struct nl_msg *msg;
+	int driver_id, i;
+
+	socket = get_nbd_socket(&driver_id);
+	nl_socket_modify_cb(socket, NL_CB_VALID, NL_CB_CUSTOM, callback, NULL);
+
+	msg = nlmsg_alloc();
+	if (!msg)
+		err("Couldn't allocate netlink message\n");
+	genlmsg_put(msg, NL_AUTO_PORT, NL_AUTO_SEQ, driver_id, 0, 0,
+		    NBD_CMD_CONNECT, 0);
+	if (index >= 0)
+		NLA_PUT_U32(msg, NBD_ATTR_INDEX, index);
+	NLA_PUT_U64(msg, NBD_ATTR_SIZE_BYTES, size64);
+	NLA_PUT_U64(msg, NBD_ATTR_BLOCK_SIZE_BYTES, blocksize);
+	NLA_PUT_U64(msg, NBD_ATTR_SERVER_FLAGS, flags);
+	if (timeout)
+		NLA_PUT_U64(msg, NBD_ATTR_TIMEOUT, timeout);
+
+	sock_attr = nla_nest_start(msg, NBD_ATTR_SOCKETS);
+	if (!sock_attr)
+		err("Couldn't nest the sockets for our connection\n");
+	for (i = 0; i < num_connects; i++) {
+		struct nlattr *sock_opt;
+		sock_opt = nla_nest_start(msg, NBD_SOCK_ITEM);
+		if (!sock_opt)
+			err("Couldn't nest the sockets for our connection\n");
+		NLA_PUT_U32(msg, NBD_SOCK_FD, sockfds[i]);
+		nla_nest_end(msg, sock_opt);
+	}
+	nla_nest_end(msg, sock_attr);
+
+	if (nl_send_sync(socket, msg) < 0)
+		err("Failed to setup device, check dmesg\n");
+	return;
+nla_put_failure:
+	err("Failed to create netlink message\n");
+}
+
+static void netlink_disconnect(char *nbddev) {
+	struct nl_sock *socket;
+	struct nl_msg *msg;
+	int driver_id;
+
+	int index = -1;
+	if (nbddev) {
+		if (sscanf(nbddev, "/dev/nbd%d", &index) != 1)
+			err("Invalid nbd device target\n");
+	}
+	if (index < 0)
+		err("Invalid nbd device target\n");
+
+	socket = get_nbd_socket(&driver_id);
+
+	msg = nlmsg_alloc();
+	if (!msg)
+		err("Couldn't allocate netlink message\n");
+	genlmsg_put(msg, NL_AUTO_PORT, NL_AUTO_SEQ, driver_id, 0, 0,
+		    NBD_CMD_DISCONNECT, 0);
+	NLA_PUT_U32(msg, NBD_ATTR_INDEX, index);
+	if (nl_send_sync(socket, msg) < 0)
+		err("Failed to disconnect device, check dmsg\n");
+	nl_socket_free(socket);
+	return;
+nla_put_failure:
+	err("Failed to create netlink message\n");
+}
+#else
+static void netlink_configure(int index, int *sockfds, int num_connects,
+			      u64 size64, int blocksize, uint16_t flags,
+			      int timeout)
+{
+}
+
+static void netlink_disconnect(char *nbddev)
+{
+}
+#endif /* HAVE_NETLINK */
+
 int check_conn(char* devname, int do_print) {
 	char buf[256];
 	char* p;
@@ -233,7 +358,6 @@ struct reply* read_reply(int sock) {
 }
 
 void ask_list(int sock) {
-	uint32_t opt;
 	uint32_t opt_server;
 	uint32_t len;
 	uint32_t lenn;
@@ -325,7 +449,7 @@ void ask_list(int sock) {
 void parse_sizes(char *buf, uint64_t *size, uint16_t *flags) {
 	memcpy(size, buf, sizeof(*size));
 	*size = ntohll(*size);
-	buf += sizeof(size);
+	buf += sizeof(*size);
 	memcpy(flags, buf, sizeof(*flags));
 	*flags = ntohs(*flags);
 
@@ -338,13 +462,24 @@ void parse_sizes(char *buf, uint64_t *size, uint16_t *flags) {
 	printf("\n");
 }
 
-void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls) {
+void send_opt_exportname(int sock, u64 *rsize64, uint16_t *flags, bool can_opt_go, char* name, uint16_t global_flags) {
+	send_request(sock, NBD_OPT_EXPORT_NAME, -1, name);
+	char b[sizeof(*flags) + sizeof(*rsize64)];
+	if(readit(sock, b, sizeof(b)) < 0 && can_opt_go) {
+		err("E: server does not support NBD_OPT_GO and dropped connection after sending NBD_OPT_EXPORT_NAME. Try -g.");
+	}
+	parse_sizes(b, rsize64, flags);
+	if(!global_flags & NBD_FLAG_NO_ZEROES) {
+		char buf[125];
+		readit(sock, buf, 124);
+	}
+}
+
+void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls, bool can_opt_go) {
 	u64 magic;
 	uint16_t tmp;
 	uint16_t global_flags;
 	char buf[256] = "\0\0\0\0\0\0\0\0\0";
-	uint32_t opt;
-	uint32_t namesize;
 	int sock = *sockp;
 
 	printf("Negotiation: ");
@@ -473,29 +608,46 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 		exit(EXIT_SUCCESS);
 	}
 
-	send_info_request(sock, NBD_OPT_GO, 0, NULL, name);
-
 	struct reply *rep = NULL;
 	
+	if(!can_opt_go) {
+		send_opt_exportname(sock, rsize64, flags, can_opt_go, name, global_flags);
+		return;
+	}
+
+	send_info_request(sock, NBD_OPT_GO, 0, NULL, name);
+
 	do {
 		if(rep != NULL) free(rep);
 		rep = read_reply(sock);
-		if(rep->reply_type & NBD_REP_FLAG_ERROR) {
-			if(rep->reply_type == NBD_REP_ERR_UNSUP) {
-				free(rep);
-				/* server doesn't support NBD_OPT_GO or NBD_OPT_INFO,
-				 * fall back to NBD_OPT_EXPORT_NAME */
-				send_request(sock, NBD_OPT_EXPORT_NAME, -1, name);
-				char b[sizeof(*flags) + sizeof(*rsize64)];
-				readit(sock, b, sizeof(b));
-				parse_sizes(b, rsize64, flags);
-				if(!(global_flags & NBD_FLAG_NO_ZEROES)) {
-					readit(sock, buf, 124);
-				}
-				return;
-			} else {
-				err("Unknown error in reply to NBD_OPT_GO; cannot continue");
-				exit(EXIT_FAILURE);
+		if(rep && (rep->reply_type & NBD_REP_FLAG_ERROR)) {
+			switch(rep->reply_type) {
+				case NBD_REP_ERR_UNSUP:
+					/* server doesn't support NBD_OPT_GO or NBD_OPT_INFO,
+					 * fall back to NBD_OPT_EXPORT_NAME */
+					send_opt_exportname(sock, rsize64, flags, can_opt_go, name, global_flags);
+					free(rep);
+					return;
+				case NBD_REP_ERR_POLICY:
+					if(rep->datasize > 0) {
+						char errstr[1024];
+						snprintf(errstr, sizeof errstr, "Connection not allowed by server policy. Server said: %s", rep->data);
+						err(errstr);
+					} else {
+						err("Connection not allowed by server policy.");
+					}
+					free(rep);
+					exit(EXIT_FAILURE);
+				default:
+					if(rep->datasize > 0) {
+						char errstr[1024];
+						snprintf(errstr, sizeof errstr, "Unknown error returned by server. Server said: %s", rep->data);
+						err(errstr);
+					} else {
+						err("Unknown error returned by server.");
+					}
+					free(rep);
+					exit(EXIT_FAILURE);
 			}
 		}
 		uint16_t info_type;
@@ -515,13 +667,13 @@ void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t n
 			case NBD_REP_ACK:
 				break;
 			default:
-				err_nonfatal("Unknown reply to NBD_OPT_GO received");
+				err_nonfatal("Unknown reply to NBD_OPT_GO received, ignoring");
 		}
 	} while(rep->reply_type != NBD_REP_ACK);
 	free(rep);
 }
 
-bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* sdp, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname) {
+bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* sdp, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname, bool *can_opt_go) {
 	int fd = open(SYSCONFDIR "/nbdtab", O_RDONLY);
 	bool retval = false;
 	if(fd < 0) {
@@ -630,6 +782,10 @@ bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** host
 			*tlshostname = strndup(loc+9, strcspn(loc+9, ","));
 			goto next;
 		}
+		if(!strncmp(loc, "no_optgo", 8)) {
+			*can_opt_go = false;
+			goto next;
+		}
 		// skip unknown options, with a warning unless they start with a '_'
 		l = strcspn(loc, ",");
 		if(*loc != '_') {
@@ -739,7 +895,11 @@ void usage(char* errmsg, ...) {
 	} else {
 		fprintf(stderr, "%s version %s\n", PROG_NAME, PACKAGE_VERSION);
 	}
+#if HAVE_NETLINK
+	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m] [-nonetlink|-L]\n");
+#else
 	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m]\n");
+#endif
 	fprintf(stderr, "Or   : nbd-client -u (with same arguments as above)\n");
 	fprintf(stderr, "Or   : nbd-client nbdX\n");
 	fprintf(stderr, "Or   : nbd-client -d nbd_device\n");
@@ -772,6 +932,12 @@ void disconnect(char* device) {
 	printf("done\n");
 }
 
+#if HAVE_NETLINK
+static const char *short_opts = "-A:b:c:C:d:H:hK:LlnN:pSst:uVx";
+#else
+static const char *short_opts = "-A:b:c:C:d:gH:hK:lnN:pSst:uVx";
+#endif
+
 int main(int argc, char *argv[]) {
 	char* port=NBD_DEFAULT_PORT;
 	int sock, nbd;
@@ -801,30 +967,38 @@ int main(int argc, char *argv[]) {
 	bool tls = false;
 	struct sigaction sa;
 	int num_connections = 1;
+	int netlink = HAVE_NETLINK;
+	int need_disconnect = 0;
+	int *sockfds;
 	struct option long_options[] = {
+		{ "cacertfile", required_argument, NULL, 'A' },
 		{ "block-size", required_argument, NULL, 'b' },
 		{ "check", required_argument, NULL, 'c' },
 		{ "connections", required_argument, NULL, 'C'},
 		{ "disconnect", required_argument, NULL, 'd' },
+		{ "certfile", required_argument, NULL, 'F' },
+		{ "no-optgo", no_argument, NULL, 'g' },
 		{ "help", no_argument, NULL, 'h' },
+		{ "tlshostname", required_argument, NULL, 'H' },
+		{ "keyfile", required_argument, NULL, 'K' },
 		{ "list", no_argument, NULL, 'l' },
-		{ "name", required_argument, NULL, 'N' },
+#if HAVE_NETLINK
+		{ "nonetlink", no_argument, NULL, 'L' },
+#endif
+		{ "systemd-mark", no_argument, NULL, 'm' },
 		{ "nofork", no_argument, NULL, 'n' },
+		{ "name", required_argument, NULL, 'N' },
 		{ "persist", no_argument, NULL, 'p' },
 		{ "sdp", no_argument, NULL, 'S' },
 		{ "swap", no_argument, NULL, 's' },
-		{ "systemd-mark", no_argument, NULL, 'm' },
 		{ "timeout", required_argument, NULL, 't' },
 		{ "unix", no_argument, NULL, 'u' },
-		{ "certfile", required_argument, NULL, 'F' },
-		{ "keyfile", required_argument, NULL, 'K' },
-		{ "cacertfile", required_argument, NULL, 'A' },
-		{ "tlshostname", required_argument, NULL, 'H' },
-		{ "enable-tls", no_argument, NULL, 'x' },
 		{ "version", no_argument, NULL, 'V' },
+		{ "enable-tls", no_argument, NULL, 'x' },
 		{ 0, 0, 0, 0 }, 
 	};
 	int i;
+	bool can_opt_go = true;
 
 	logging(MY_NAME);
 
@@ -832,7 +1006,7 @@ int main(int argc, char *argv[]) {
         tlssession_init();
 #endif
 
-	while((c=getopt_long_only(argc, argv, "-b:c:d:hlnN:pSst:uVC:K:A:H:x", long_options, NULL))>=0) {
+	while((c=getopt_long_only(argc, argv, short_opts, long_options, NULL))>=0) {
 		switch(c) {
 		case 1:
 			// non-option argument
@@ -885,8 +1059,12 @@ int main(int argc, char *argv[]) {
 			num_connections = (int)strtol(optarg, NULL, 0);
 			break;
 		case 'd':
-			disconnect(optarg);
-			exit(EXIT_SUCCESS);
+			need_disconnect = 1;
+			nbddev = strdup(optarg);
+			break;
+		case 'g':
+			can_opt_go = false;
+			break;
 		case 'h':
 			usage(NULL);
 			exit(EXIT_SUCCESS);
@@ -895,6 +1073,11 @@ int main(int argc, char *argv[]) {
 			opts |= NBDC_DO_LIST;
 			nbddev="";
 			break;
+#if HAVE_NETLINK
+		case 'L':
+			netlink = 0;
+			break;
+#endif
 		case 'm':
 			argv[0][0] = '@';
 			break;
@@ -953,6 +1136,13 @@ int main(int argc, char *argv[]) {
 		}
 	}
 
+	if (need_disconnect) {
+		if (netlink)
+			netlink_disconnect(nbddev);
+		else
+			disconnect(nbddev);
+		exit(EXIT_SUCCESS);
+	}
 #ifdef __ANDROID__
   if (swap)
     err("swap option unsupported on Android because mlockall is unsupported.");
@@ -960,11 +1150,11 @@ int main(int argc, char *argv[]) {
 	if(hostname) {
 		if((!name || !nbddev) && !(opts & NBDC_DO_LIST)) {
 			if(!strncmp(hostname, "nbd", 3) || !strncmp(hostname, "/dev/nbd", 8)) {
-				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &sdp, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &hostname)) {
+				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &sdp, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &hostname, &can_opt_go)) {
 					usage("no valid configuration for specified device found", hostname);
 					exit(EXIT_FAILURE);
 				}
-			} else {
+			} else if (!netlink) {
 				usage("not enough information specified, and argument didn't look like an nbd device");
 				exit(EXIT_FAILURE);
 			}
@@ -984,16 +1174,25 @@ int main(int argc, char *argv[]) {
         if (!tlshostname && hostname)
                 tlshostname = strdup(hostname);
 
+	if (netlink)
+		nofork = 1;
+
 	if(strlen(name)==0 && !(opts & NBDC_DO_LIST)) {
 		printf("Warning: the oldstyle protocol is no longer supported.\nThis method now uses the newstyle protocol with a default export\n");
 	}
 
-	if(!(opts & NBDC_DO_LIST)) {
+	if(!(opts & NBDC_DO_LIST) && !netlink) {
 		nbd = open(nbddev, O_RDWR);
 		if (nbd < 0)
 			err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
 	}
 
+	if (netlink) {
+		sockfds = malloc(sizeof(int) * num_connections);
+		if (!sockfds)
+			err("Cannot allocate the socket fd's array");
+	}
+
 	for (i = 0; i < num_connections; i++) {
 		if (b_unix)
 			sock = openunix(hostname);
@@ -1002,7 +1201,12 @@ int main(int argc, char *argv[]) {
 		if (sock < 0)
 			exit(EXIT_FAILURE);
 
-		negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls);
+		negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, can_opt_go);
+		if (netlink) {
+			sockfds[i] = sock;
+			continue;
+		}
+
 		if (i == 0) {
 			setsizes(nbd, size64, blocksize, flags);
 			set_timeout(nbd, timeout);
@@ -1019,6 +1223,16 @@ int main(int argc, char *argv[]) {
 		}
 	}
 
+	if (netlink) {
+		int index = -1;
+		if (nbddev) {
+			if (sscanf(nbddev, "/dev/nbd%d", &index) != 1)
+				err("Invalid nbd device target\n");
+		}
+		netlink_configure(index, sockfds, num_connections,
+				  size64, blocksize, flags, timeout);
+		return 0;
+	}
 	/* Go daemon */
 	
 #ifndef NOFORK
@@ -1077,7 +1291,7 @@ int main(int argc, char *argv[]) {
 
 		if (ioctl(nbd, NBD_DO_IT) < 0) {
 			int error = errno;
-			fprintf(stderr, "nbd,%d: Kernel call returned: %d", main_pid, error);
+			fprintf(stderr, "nbd,%d: Kernel call returned: %s\n", main_pid, strerror(errno));
 			if(error==EBADR) {
 				/* The user probably did 'nbd-client -d' on us.
 				 * quit */
@@ -1101,7 +1315,7 @@ int main(int argc, char *argv[]) {
 					nbd = open(nbddev, O_RDWR);
 					if (nbd < 0)
 						err("Cannot open NBD: %m");
-					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls);
+					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls, can_opt_go);
 					if (size64 != new_size) {
 						err("Size of the device changed. Bye");
 					}
@@ -1116,7 +1330,7 @@ int main(int argc, char *argv[]) {
 			/* We're on 2.4. It's not clearly defined what exactly
 			 * happened at this point. Probably best to quit, now
 			 */
-			fprintf(stderr, "Kernel call returned.");
+			fprintf(stderr, "Kernel call returned.\n");
 			cont=0;
 		}
 	} while(cont);
